#lang racket

;; Slippy: lambda letrec if set! begin quote

(define ns (make-base-namespace))
(struct ev (e ρ σ κ))
(struct ko (φ v σ κ))
(struct ap (v vs ρ σ κ))
(struct letk (a es ρ))
(struct setk (x ρ))
(struct ifk (e1 e2 ρ))
(struct seqk (es ρ))
(struct randk (xs vs ρ))
(struct haltk ())
(struct clo (λ ρ))
(struct lam (x es))
(define (eval-seq es ρ σ κ)
 (match es
   ((list e) (ev e ρ σ κ))
   ((cons e es) (ev e ρ σ (cons (seqk es ρ) κ)))))
(define (step state)
 (match state
   ((ev (and x (? symbol?)) ρ σ (cons φ κ))
    (ko φ (match (assoc x ρ)
            ((cons _ a) (cdr (assoc a σ)))
            (_ (eval x ns))) σ κ))
   ((ev `(lambda ,x ,es ...) ρ σ (cons φ κ))
    (ko φ (clo (lam x es) ρ) σ κ))
   ((ev `(quote ,e) ρ σ (cons φ κ))
    (ko φ e σ κ))
   ((ev `(if ,e ,e1 ,e2) ρ σ κ)
    (ev e ρ σ (cons (ifk e1 e2 ρ) κ)))
   ((ev `(letrec ((,x ,e)) ,es ...) ρ σ κ)
    (let* ((fresh (gensym))
           (ρ* (cons (cons x fresh) ρ)))
      (ev e ρ* σ (cons (letk fresh es ρ*) κ))))
   ((ev `(set! ,x ,e) ρ σ κ)
    (ev e ρ σ (cons (setk x ρ) κ)))
   ((ev `(begin ,es ...) ρ σ κ)
    (eval-seq es ρ σ κ))
   ((ev `(,rator . ,rands) ρ σ κ)
    (ev rator ρ σ (cons (randk rands '() ρ) κ)))
   ((ev e ρ σ (cons φ κ))
    (ko φ e σ κ))
   ((ko (letk a es ρ) v σ κ)
    (let ((σ* (cons (cons a v) σ)))
      (eval-seq es ρ σ* κ)))
   ((ko (setk x ρ) v σ (cons φ κ))
    (match (assoc x ρ)
      ((cons name a) (ko φ v (cons (cons a v) σ) κ))))
   ((ko (randk '() vs ρ) v σ κ)
    (let ((vs (reverse (cons v vs))))
      (ap (car vs) (cdr vs) ρ σ κ)))
   ((ap (clo (lam x es) ρ*) rands ρ σ κ)
    (let loop ((x x) (rands rands) (ρ ρ*) (σ σ))
      (match x
        ('() (eval-seq es ρ σ κ))
        ((cons x xs)
         (let ((fresh (gensym)))
           (loop xs (cdr rands)
                 (cons (cons x fresh) ρ)
                 (cons (cons fresh (car rands)) σ))))
        ((and x (? symbol?))
         (let ((fresh (gensym)))
           (eval-seq es
                     (cons (cons x fresh) ρ)
                     (cons (cons fresh rands) σ) κ))))))
   ((ap rator rands _ σ (cons φ κ))
    (ko φ (apply rator rands) σ κ))
   ((ko (randk rands vs ρ) v σ κ)
    (ev (car rands) ρ σ (cons (randk (cdr rands) (cons v vs) ρ) κ)))
   ((ko (ifk _ e2 ρ) #f σ κ)
    (ev e2 ρ σ κ))
   ((ko (ifk e1 _ ρ) _ σ κ)
    (ev e1 ρ σ κ))
   ((ko (seqk (list e) ρ) _ σ κ)
    (ev e ρ σ κ))
   ((ko (seqk (cons e exps) ρ) _ σ κ)
    (ev e ρ σ (cons (seqk exps ρ) κ)))
   ((ko (haltk) v _ _)
    #f)))


(define (inject e)
 (ev e '() '() `(,(haltk))))
(define (run s)
 (let ((next (step s)))
   (if next
       (run next)
       s)))
(define (extract s)
 (match s
   ((ko (haltk) v _ _) v)
   (_ 'error)))
(define (state-eval e)
 (extract (run (inject e))))