#lang racket

(define ns (make-base-namespace))

(struct ev (e ρ σ κ))
(struct ko (φ v ρ* σ κ))
(struct defk (fresh))
(struct setk (x))
(struct ifk (e))
(struct ifek (e1 e2))
(struct seqk (es))
(struct randk (rands vs))
(struct thkk (ρ))
(struct repk ())
(struct clo (λ ρ))
(struct lam (x es))

(define (eval-seq es ρ σ κ)
  (match es
    ((list e) (ev e ρ σ κ))
    ((cons e es) (ev e ρ σ (cons (seqk es) κ)))))

(define (step state)
  (match state 
    ((ev (and x (? symbol?)) ρ σ (cons φ κ))
     (ko φ (match (assoc x ρ)
             ((cons _ a) (cdr (assoc a σ)))
             (_ (eval x ns))) ρ σ κ))
    
    ((ev `(lambda ,x ,es ...) ρ σ (cons φ κ))
     (let ((clo (clo (lam x es) ρ)))
       (ko φ clo ρ σ κ)))
    
    ((ev `(quote ,e) ρ σ (cons φ κ))
     (ko φ e ρ σ))    ; <--- e should be cloned
    
    ((ev `(if ,e ,e1) ρ σ κ)
     (ev e ρ σ (cons (ifk e1) κ)))
    
    ((ev `(if ,e ,e1 ,e2) ρ σ κ)
     (ev e ρ σ (cons (ifek e1 e2) κ)))
    
    ((ev `(define ,(cons x xs) ,es ...) ρ σ (cons φ κ))
     (let* ((fresh (gensym))
            (ρ* (cons (cons x fresh) ρ))
            (clo (clo (lam xs es) ρ*))
            (σ* (cons (cons fresh clo) σ)))
       (ko φ clo ρ* σ* κ)))
    
    ((ev `(define ,x ,e) ρ σ κ)
     (let* ((fresh (gensym))
            (ρ* (cons (cons x fresh) ρ)))
       (ev e ρ* σ (cons (defk fresh) κ))))
    
    ((ev `(set! ,x ,e) ρ σ κ)
     (ev e ρ σ (cons (setk x) κ)))
    
    ((ev `(begin ,es ...) ρ σ κ)
     (eval-seq es ρ σ κ))
    
    ((ev `(,rator . ,rands) ρ σ κ)
     (ev rator ρ σ (cons (randk rands '()) κ)))
    
    ((ev e ρ σ (cons φ κ))
     (ko φ e ρ σ κ))
    
    ((ko (defk fresh) v ρ σ (cons φ κ))
     (let ((σ* (cons (cons fresh v) σ)))
      (ko φ v ρ σ* κ)))
    
    ((ko (setk x) v ρ σ (cons φ κ))
     (match (assoc x ρ)
       ((cons name a) (ko φ v ρ (cons (cons a v) σ) κ))))
    
    ((ko (randk '() vs) v ρ σ (cons φ κ))
     (match (reverse (cons v vs))
       ((cons (clo (lam x es) ρ*) rands)
        (let loop ((x x) (rands rands) (ρ ρ*) (σ σ))
          (match x
            ('() (eval-seq es ρ σ (cons (thkk ρ) (cons φ κ))))
            ((cons x xs)
             (let ((fresh (gensym)))
               (loop xs (cdr rands)
                     (cons (cons x fresh) ρ)
                     (cons (cons fresh (car rands)) σ))))
            ((and x (? symbol?))
             (let* ((fresh (gensym))
                    (ρ* (cons (cons x fresh) ρ))
                    (σ* (cons (cons fresh rands) σ)))   
               (eval-seq es ρ* σ* (cons (thkk ρ) (cons φ κ))))))))
        ((cons rator rands)
         (ko φ (apply rator rands) ρ σ κ))))
    
    ((ko (randk rands vs) v ρ σ κ)
     (ev (car rands) ρ σ (cons (randk (cdr rands) (cons v vs)) κ)))
    
    ((ko (ifk _) #f ρ σ (cons φ κ))
     (ko φ '() ρ σ κ))
    
    ((ko (ifk e) _ ρ σ κ)
     (ev e ρ σ (cons (thkk ρ) κ)))
    
    ((ko (ifek _ e2) #f ρ σ κ)
     (ev e2 ρ σ (cons (thkk ρ) κ)))
    
    ((ko (ifek e1 _) _ ρ σ κ)
     (ev e1 ρ σ (cons (thkk ρ) κ)))
    
    ((ko (seqk (list e)) _ ρ σ κ)
     (ev e ρ σ κ))
    
    ((ko (seqk (cons e exps) ) _ ρ σ κ)
     (ev e ρ σ (cons (seqk exps) κ)))
    
    ((ko (thkk ρ) v _ σ (cons φ κ))
     (ko φ v ρ σ κ))
    
    ((ko (repk) v ρ σ _)
     (display v) (newline) (display "> ") (ev (read) ρ σ (list (repk))))))

(let loop ((state (ko (repk) "Small-Step Slip" '() '() '())))
  (loop (step state)))